from __future__ import annotations

from typing import Any

import pytest

import narwhals as nw
from tests.utils import assert_equal_data

pytest.importorskip("pandas", minversion="3.0.0")
pytest.importorskip("pyarrow")

import numpy as np
import pandas as pd
import pyarrow as pa

STRING_DTYPE_NAN = pd.StringDtype("pyarrow", na_value=np.nan)  # type: ignore[call-arg]
STRING_DTYPE_NA = pd.StringDtype("pyarrow", na_value=pd.NA)  # type: ignore[call-arg]
STRING_DTYPE_PYTHON_NAN = pd.StringDtype("python", na_value=np.nan)  # type: ignore[call-arg]
STRING_DTYPE_PYTHON_NA = pd.StringDtype("python", na_value=pd.NA)  # type: ignore[call-arg]


@pytest.mark.parametrize(
    ("left_dtype", "right_dtype", "result_dtype"),
    [
        (STRING_DTYPE_NAN, STRING_DTYPE_NAN, STRING_DTYPE_NAN),
        (STRING_DTYPE_NAN, STRING_DTYPE_NA, STRING_DTYPE_NAN),
        (STRING_DTYPE_NAN, pd.ArrowDtype(pa.string()), STRING_DTYPE_NAN),
        (STRING_DTYPE_NAN, pd.ArrowDtype(pa.large_string()), STRING_DTYPE_NAN),
        (STRING_DTYPE_NAN, STRING_DTYPE_PYTHON_NAN, STRING_DTYPE_NAN),
        (STRING_DTYPE_NAN, STRING_DTYPE_PYTHON_NA, STRING_DTYPE_NAN),
        (STRING_DTYPE_NA, STRING_DTYPE_NAN, STRING_DTYPE_NA),
        (STRING_DTYPE_NA, STRING_DTYPE_NA, STRING_DTYPE_NA),
        (STRING_DTYPE_NA, pd.ArrowDtype(pa.string()), STRING_DTYPE_NA),
        (STRING_DTYPE_NA, pd.ArrowDtype(pa.large_string()), STRING_DTYPE_NA),
        (STRING_DTYPE_NA, STRING_DTYPE_PYTHON_NAN, STRING_DTYPE_NA),
        (STRING_DTYPE_NA, STRING_DTYPE_PYTHON_NA, STRING_DTYPE_NA),
        (pd.ArrowDtype(pa.string()), STRING_DTYPE_NAN, pd.ArrowDtype(pa.large_string())),
        (pd.ArrowDtype(pa.string()), STRING_DTYPE_NA, pd.ArrowDtype(pa.large_string())),
        (
            pd.ArrowDtype(pa.string()),
            pd.ArrowDtype(pa.string()),
            pd.ArrowDtype(pa.string()),
        ),
        (
            pd.ArrowDtype(pa.string()),
            pd.ArrowDtype(pa.large_string()),
            pd.ArrowDtype(pa.large_string()),
        ),
        (
            pd.ArrowDtype(pa.large_string()),
            STRING_DTYPE_NAN,
            pd.ArrowDtype(pa.large_string()),
        ),
        (
            pd.ArrowDtype(pa.large_string()),
            STRING_DTYPE_NA,
            pd.ArrowDtype(pa.large_string()),
        ),
        (
            pd.ArrowDtype(pa.large_string()),
            pd.ArrowDtype(pa.string()),
            pd.ArrowDtype(pa.large_string()),
        ),
        (
            pd.ArrowDtype(pa.large_string()),
            pd.ArrowDtype(pa.large_string()),
            pd.ArrowDtype(pa.large_string()),
        ),
        (STRING_DTYPE_PYTHON_NAN, STRING_DTYPE_PYTHON_NAN, STRING_DTYPE_PYTHON_NAN),
        (STRING_DTYPE_PYTHON_NAN, STRING_DTYPE_PYTHON_NA, STRING_DTYPE_PYTHON_NA),
        (STRING_DTYPE_PYTHON_NAN, STRING_DTYPE_NAN, STRING_DTYPE_NAN),
        (STRING_DTYPE_PYTHON_NAN, STRING_DTYPE_NA, STRING_DTYPE_NA),
        (STRING_DTYPE_PYTHON_NAN, pd.ArrowDtype(pa.string()), pd.ArrowDtype(pa.string())),
        (
            STRING_DTYPE_PYTHON_NAN,
            pd.ArrowDtype(pa.large_string()),
            pd.ArrowDtype(pa.large_string()),
        ),
        (STRING_DTYPE_PYTHON_NA, STRING_DTYPE_PYTHON_NAN, STRING_DTYPE_PYTHON_NA),
        (STRING_DTYPE_PYTHON_NA, STRING_DTYPE_PYTHON_NA, STRING_DTYPE_PYTHON_NA),
        (STRING_DTYPE_PYTHON_NA, STRING_DTYPE_NAN, STRING_DTYPE_PYTHON_NA),
        (STRING_DTYPE_PYTHON_NA, STRING_DTYPE_NA, STRING_DTYPE_NA),
        (STRING_DTYPE_PYTHON_NA, pd.ArrowDtype(pa.string()), pd.ArrowDtype(pa.string())),
        (
            STRING_DTYPE_PYTHON_NA,
            pd.ArrowDtype(pa.large_string()),
            pd.ArrowDtype(pa.large_string()),
        ),
        (pd.ArrowDtype(pa.string()), STRING_DTYPE_PYTHON_NAN, pd.ArrowDtype(pa.string())),
        (pd.ArrowDtype(pa.string()), STRING_DTYPE_PYTHON_NA, pd.ArrowDtype(pa.string())),
        (
            pd.ArrowDtype(pa.large_string()),
            STRING_DTYPE_PYTHON_NAN,
            pd.ArrowDtype(pa.large_string()),
        ),
        (
            pd.ArrowDtype(pa.large_string()),
            STRING_DTYPE_PYTHON_NA,
            pd.ArrowDtype(pa.large_string()),
        ),
    ],
)
def test_pandas_str_types(left_dtype: Any, right_dtype: Any, result_dtype: Any) -> None:
    df = pd.DataFrame({"fruit": ["apple", "banana"]}, dtype=left_dtype)
    df["new_str_col"] = "!"
    df["new_str_col"] = df["new_str_col"].astype(right_dtype)  # pyrefly: ignore[missing-attribute] https://github.com/facebook/pyrefly/issues/3299
    res = nw.from_native(df).with_columns(
        concat_col=nw.concat_str([nw.col("fruit"), nw.col("new_str_col")])
    )
    expected = {
        "fruit": ["apple", "banana"],
        "new_str_col": ["!", "!"],
        "concat_col": ["apple!", "banana!"],
    }
    assert_equal_data(res, expected)
    assert res.to_native()["concat_col"].dtype == result_dtype



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import narwhals as nw
from tests.utils import PANDAS_VERSION, POLARS_VERSION, PYARROW_VERSION

if TYPE_CHECKING:
    from tests.utils import Constructor, ConstructorEager

data = {"a": ["12:34:56"]}


def requires_time_support(
    request: pytest.FixtureRequest, constructor: Constructor | ConstructorEager
) -> None:
    """Enforce Time dtype test expectations for dataframe backends.

    Skip or mark tests as expected failures depending on backend capabilities,
    version, and pyarrow availability when testing Time dtype support.
    """
    if constructor.__name__.startswith(("pandas", "modin")):
        if PANDAS_VERSION < (2, 2, 0):
            pytest.skip(
                "pandas < 2.2.0 has no pyarrow dtype support (and therefore does not support the Time dtype)"
            )

        if PYARROW_VERSION == (0, 0, 0):
            pytest.skip("pandas requires pyarrow for the Time dtype")

    if "pyspark" in str(constructor) or "dask" in str(constructor):
        request.applymarker(
            pytest.mark.xfail(reason="backend does not support Time dtype")
        )


def test_to_time(request: pytest.FixtureRequest, constructor: Constructor) -> None:
    requires_time_support(request, constructor)

    expected = "12:34:56"

    result = (
        nw.from_native(constructor(data))
        .lazy()
        .select(b=nw.col("a").str.to_time(format="%H:%M:%S"))
        .collect()
    )
    assert isinstance(result.collect_schema()["b"], nw.Time)
    assert str(result.item(row=0, column="b")) == expected


def test_to_time_series(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    requires_time_support(request, constructor_eager)

    expected = "12:34:56.000000000" if "cudf" in str(constructor_eager) else "12:34:56"
    result = nw.from_native(constructor_eager(data), eager_only=True)["a"].str.to_time(
        format="%H:%M:%S"
    )

    assert isinstance(result.dtype, nw.Time)
    assert str(result.item(0)) == expected


@pytest.mark.parametrize(
    ("data", "expected"),
    [({"a": ["12:34:56"]}, "12:34:56"), ({"a": ["12:34"]}, "12:34:00")],
)
def test_to_time_infer_fmt(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    data: dict[str, list[str]],
    expected: str,
) -> None:
    requires_time_support(request, constructor)

    if (
        "polars" in str(constructor)
        and POLARS_VERSION < (1, 30)
        and data["a"][0].count(":") < 2
    ):  # pragma: no cover
        request.applymarker(
            pytest.mark.xfail(reason="Polars<1.30 cannot auto-infer the HH:MM format")
        )

    result = (
        nw.from_native(constructor(data))
        .lazy()
        .select(b=nw.col("a").str.to_time())
        .collect()
    )
    assert str(result.item(row=0, column="b")) == expected
    assert isinstance(result.collect_schema()["b"], nw.Time)


@pytest.mark.parametrize(
    ("data", "expected"),
    [({"a": ["12:34:56"]}, "12:34:56"), ({"a": ["12:34"]}, "12:34:00")],
)
def test_to_time_series_infer_fmt(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    data: dict[str, list[str]],
    expected: str,
) -> None:
    requires_time_support(request, constructor_eager)

    if (
        "polars" in str(constructor_eager)
        and POLARS_VERSION < (1, 30)
        and data["a"][0].count(":") < 2
    ):  # pragma: no cover
        request.applymarker(
            pytest.mark.xfail(reason="Polars<1.30 cannot auto-infer the HH:MM format")
        )

    result = nw.from_native(constructor_eager(data), eager_only=True)["a"].str.to_time()
    assert str(result.item(0)) == expected
    assert isinstance(result.dtype, nw.Time)



from __future__ import annotations

import pytest

import narwhals as nw
from tests.utils import Constructor, ConstructorEager, assert_equal_data

EXPR_UNSUPPORTED = ("dask", "pyarrow", "pandas", "modin", "cudf")
data = {"a": ["fdas", "edfas"], "prefix_suffix": ["fda", "fas"]}


@pytest.mark.parametrize(
    ("suffix", "expected"), [("das", [True, False]), (nw.lit("das"), [True, False])]
)
def test_ends_with(
    constructor: Constructor, suffix: str | nw.Expr, expected: list[bool]
) -> None:
    df = nw.from_native(constructor(data))
    result = df.select(nw.col("a").str.ends_with(suffix))
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    ("suffix", "expected"), [(nw.col("prefix_suffix"), [False, True])]
)
def test_ends_with_multi(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    suffix: nw.Expr,
    expected: list[bool],
) -> None:
    if any(x in str(constructor) for x in EXPR_UNSUPPORTED):
        request.applymarker(pytest.mark.xfail(reason="Not supported", raises=TypeError))
    df = nw.from_native(constructor(data))
    result = df.select(nw.col("a").str.ends_with(suffix))
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(("suffix", "expected"), [("das", [True, False])])
def test_ends_with_series(
    constructor_eager: ConstructorEager, suffix: str, expected: list[bool]
) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df.select(df["a"].str.ends_with(suffix))
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(("suffix", "expected"), [(data["prefix_suffix"], [False, True])])
def test_ends_with_series_multi(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    suffix: str | nw.Expr,
    expected: list[bool],
) -> None:
    if any(x in str(constructor_eager) for x in EXPR_UNSUPPORTED):
        request.applymarker(pytest.mark.xfail(reason="Not supported", raises=TypeError))

    df = nw.from_native(constructor_eager(data), eager_only=True)
    suffix_series = nw.from_native(constructor_eager({"a": suffix}), eager_only=True)["a"]
    result = df.select(df["a"].str.ends_with(suffix_series))

    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    ("prefix", "expected"), [("fda", [True, False]), (nw.lit("fda"), [True, False])]
)
def test_starts_with(
    constructor: Constructor, prefix: str | nw.Expr, expected: list[bool]
) -> None:
    df = nw.from_native(constructor(data)).lazy()
    result = df.select(nw.col("a").str.starts_with(prefix))
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    ("prefix", "expected"), [(nw.col("prefix_suffix"), [True, False])]
)
def test_starts_with_multi(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    prefix: nw.Expr,
    expected: list[bool],
) -> None:
    if any(x in str(constructor) for x in EXPR_UNSUPPORTED):
        request.applymarker(pytest.mark.xfail(reason="Not supported", raises=TypeError))

    df = nw.from_native(constructor(data)).lazy()
    result = df.select(nw.col("a").str.starts_with(prefix))
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(("prefix", "expected"), [("fda", [True, False])])
def test_starts_with_series(
    constructor_eager: ConstructorEager, prefix: str, expected: list[bool]
) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df.select(df["a"].str.starts_with(prefix))
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(("prefix", "expected"), [(data["prefix_suffix"], [True, False])])
def test_starts_with_series_multi(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    prefix: list[str],
    expected: list[bool],
) -> None:
    if any(x in str(constructor_eager) for x in EXPR_UNSUPPORTED):
        request.applymarker(pytest.mark.xfail(reason="Not supported", raises=TypeError))

    prefix_series = nw.from_native(constructor_eager({"a": prefix}), eager_only=True)["a"]
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df.select(df["a"].str.starts_with(prefix_series))
    assert_equal_data(result, {"a": expected})
